#!/usr/bin/python2.7
#
# Univention DigiOnline WebWeaver integration
"""
Setup Univention Directory Listener modules for DigiOnline WebWeaver
integration as configured through the UCR variables webweaver/....
"""
#
# Copyright (C) 2014 Univention GmbH <http://www.univention.de/>
# Copyright (C) 2014 DigiOnline GmbH <http://www.digionline.de/>
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import os
import re
import sys
from json import dump, JSONEncoder
from errno import ENOENT
from subprocess import call
from optparse import OptionParser, SUPPRESS_HELP
from univention.config_registry import ConfigRegistry
import logging
from pwd import getpwnam

RE_UCR = re.compile(r'^webweaver/([^/]+)/(ServerLogin|ServerPasswd|ServerURL|error_email|mapping/([^/]+))$')
RE_SUBST = re.compile(r'@%@(.+?)@%@')
CONFIGURATION = '/etc/univention-digionline-webweaver.json'
TEMPLATE = '/usr/share/univention-digionline-webweaver/univention-digionline-webweaver.py'
LISTENER_DIR = '/usr/lib/univention-directory-listener/system'
STATUS_DIR = '/var/lib/univention-directory-listener/handlers'
PREFIX = 'webweaver_'
SUFFIX = '.py'
SPOOL_DIR = '/var/spool/univention-digionline-webweaver'

is_true = lambda value, ucr=ConfigRegistry(): ucr.is_true(value=value)


def handler(config_registry, changes):
	"""
	Called as a UCR module to do the update.
	'config_registry' is an instance of univention.config_registry.ConfigRegistry.
	'changes' is a dictionary mapping the name of changed variables to a 2-tuple (old-value, new-value).
	"""
	logging.info('Handling changes: %r', changes)

	for (var, change) in changes.iteritems():
		match = RE_UCR.match(var)
		if match:
			break
	else:
		return

	schools = read_all(config_registry)
	write_config(schools)
	listener_instances(schools)


class WebWeaver(object):
	def __init__(self):
		self.parser = None
		self.options = None
		self.args = None

	def main(self):
		"""
		Command line interface.
		"""
		self.parse_args()
		logging.basicConfig(
			stream=sys.stderr,
			level=logging.DEBUG if self.options.verbose else logging.INFO)
		self.options.mode()

	def parse_args(self):
		usage = '%prog [options] command ...'
		description = sys.modules[__name__].__doc__

		parser = OptionParser(usage=usage, description=description)
		parser.set_defaults(mode=self.run_missing)
		parser.add_option(
			'-r', '--resync',
			dest='mode', action='store_const', const=self.run_resync,
			help='Re-run listener module for school')
		parser.add_option(
			'-u', '--upgrade',
			dest='mode', action='store_const', const=self.run_upgrade,
			help='Upgrade all listener modules to latest version')
		parser.add_option(
			'-l', '--list',
			dest='mode', action='store_const', const=self.run_list,
			help='List status of schools')
		parser.add_option(
			'--test',
			dest='mode', action='store_const', const=self.run_tests,
			help=SUPPRESS_HELP)
		parser.add_option(
			'-a', '--all',
			dest='all', action='store_true',
			help='Run action on all schools')
		parser.add_option(
			'-v', '--verbose',
			dest='verbose', action='count', default=0,
			help='Increase log verbosity')

		self.options, self.args = parser.parse_args()
		self.parser = parser

	def get_schools(self):
		ucr = ConfigRegistry()
		ucr.load()

		schools = read_all(ucr)
		if not self.options.all:
			schools = dict(
				(name, school)
				for (name, school) in schools.iteritems()
				if school.sync_school in self.args
			)
			if not schools:
				self.parser.error('No matching schools found.')

		return schools

	def run_missing(self):
		self.parser.error('Missing command.')

	def run_tests(self):
		logging.info('Running tests...')
		import doctest
		print doctest.testmod()

	def run_resync(self):
		logging.info('Running resync...')
		schools = self.get_schools()
		cmd = [
			'/usr/sbin/univention-directory-listener-ctrl',
			'resync',
		] + [PREFIX + _.safe_name for _ in schools.itervalues()]
		logging.debug('Running %r...', cmd)
		call(cmd)

	def run_upgrade(self):
		logging.info('Running upgrade...')
		schools = self.get_schools()
		for school in schools.itervalues():
			if school.valid():
				listener_create(school)
		if schools:
			listener_restart()

	def run_list(self):
		logging.info('List status...')
		schools = self.get_schools()
		data = [
			(
				'complete' if school.valid() else 'incomplete',
				'exists' if os.path.exists(school.listener_filename) else 'missing',
				'%7d' % len(os.listdir(school.spool_dir)) if os.path.exists(school.spool_dir) else '-',
				school.sync_school,
			)
			for school in schools.values()
		]
		data.sort()
		fmt = ' | '.join(('%10s', '%8s', '%7s', '%4s'))
		print fmt % ('SETTING', 'LISTENER', 'PENDING', 'NAME')
		print (fmt % ('', '', '', '')).replace(' ', '-')
		for line in data:
			print fmt % line


class School(object):
	SERVER_URL = 'https://www.wwschool.de/soap.php?wsdl'

	MAPPING = {
		'lehrer': 0,
		'schueler': 1,
		'mitarbeiter': 2,
		'l_und_m': 3,
		'klassen': 19,
		'gruppen': 18,
	}

	def __init__(self, name):
		self.sync_school = name
		self.ServerLogin = None
		self.ServerPasswd = None
		self.ServerURL = None
		self.error_email = None
		self.mapping = {}

	def __str__(self):
		return self.sync_school

	def fill_missing(self, ucr):
		"""
		Fetch unchanged values from UCR.
		"""
		self._default(ucr, 'ServerLogin')
		self._default(ucr, 'ServerPasswd')
		self._default(ucr, 'ServerURL', self.SERVER_URL)
		self._default(ucr, 'error_email')

	def _default(self, ucr, key, default=None):
		if not getattr(self, key):
			value = ucr.get('webweaver/%s/%s' % (self.sync_school, key), default)
			setattr(self, key, value)

	def valid(self):
		return all((self.sync_school, self.ServerLogin, self.ServerPasswd, self.ServerURL))

	def json(self):
		"""
		Dump configuration as JSON.

		>>> json = School('UCS_WW3EE_Demo').json()
		>>> json['sync_school']
		'UCS_WW3EE_Demo'
		"""
		result = dict(self.mapping or self.MAPPING)
		result.update({
			'ServerURL': self.ServerURL,
			'ServerLogin': self.ServerLogin,
			'ServerPasswd': self.ServerPasswd,
			'sync_school': self.sync_school,
			'error_email': self.ServerLogin if is_true(self.error_email) else '',
			'ExportDirectory': self.spool_dir,
		})
		return result

	def __setattr__(self, key, value):
		if key.startswith('mapping/'):
			self.mapping[key[len('mapping/'):]] = value
		else:
			object.__setattr__(self, key, value)

	@property
	def safe_name(self):
		"""
		Return safe variant of name for use in file systems.

		>>> School('Test this/not').safe_name
		'Test_this_not'
		"""
		return ''.join(_ if _.isalnum() else '_' for _ in self.sync_school)

	@property
	def listener_filename(self):
		"""
		Path name of the listener module.
		"""
		return os.path.join(LISTENER_DIR, PREFIX + self.safe_name + SUFFIX)

	@property
	def status_filename(self):
		"""
		Path name of the listener status file.
		"""
		return os.path.join(STATUS_DIR, PREFIX + self.safe_name)

	@property
	def spool_dir(self):
		"""
		Path name of the spool directory.
		"""
		return os.path.join(SPOOL_DIR, self.safe_name)


class SchoolJson(JSONEncoder):
	def default(self, o):
		if isinstance(o, School):
			return o.json()
		return super(SchoolJson, self).default(o)


def process_changes(changes):
	"""
	Process changes UCR variables.

	>>> from hashlib import md5
	>>> schools = process_changes({
	...  'webweaver/UCS_WW3EE_Demo/ServerLogin': (None, 'synchronisation@ucs.wwschool.de'),
	...  'webweaver/UCS_WW3EE_Demo/ServerPasswd': (None, md5('mj38zsASda')),
	...  'webweaver/UCS_WW3EE_Demo/error_email': (None, 'true'),
	... })
	>>> schools['UCS_WW3EE_Demo'].ServerLogin
	'synchronisation@ucs.wwschool.de'
	"""
	schools = {}
	for (var, change) in changes.iteritems():
		if not change:
			continue
		(old, new) = change
		if not new:
			continue
		match = RE_UCR.match(var)
		if not match:
			continue
		school_name, key, subkey = match.groups()
		school = schools.setdefault(school_name, School(school_name))
		setattr(school, key, new)
	return schools


def fill_missing(schools, ucr):
	"""
	Fetch missing values from UCR.

	>>> from hashlib import md5
	>>> schools = {'UCS_WW3EE_Demo': School('UCS_WW3EE_Demo')}
	>>> fill_missing(schools, {
	...  'webweaver/UCS_WW3EE_Demo/ServerLogin': 'synchronisation@ucs.wwschool.de',
	...  'webweaver/UCS_WW3EE_Demo/ServerPasswd': md5('mj38zsASda'),
	...  'webweaver/UCS_WW3EE_Demo/error_email': 'true',
	... })
	>>> schools['UCS_WW3EE_Demo'].ServerURL
	'https://www.wwschool.de/soap.php?wsdl'
	"""
	for school in schools.itervalues():
		school.fill_missing(ucr)


def write_config(schools):
	logging.info('Writing new configuration file "%s"...', CONFIGURATION)
	ent = getpwnam('listener')
	tmp = CONFIGURATION + '.tmp'
	configured = [_ for _ in schools.itervalues() if _.valid()]
	with open(tmp, 'wb') as config:
		os.fchmod(config.fileno(), 0600)
		os.fchown(config.fileno(), ent.pw_uid, ent.pw_gid)
		dump(configured, config, cls=SchoolJson)
	os.rename(tmp, CONFIGURATION)


def listener_instances(schools):
	restart = False
	existing = set([
		_[len(PREFIX):-len(SUFFIX)]
		for _ in os.listdir(LISTENER_DIR)
		if _.startswith(PREFIX) and _.endswith(SUFFIX)
	])
	logging.debug('existing=%r', existing)
	configured = dict(
		(name, school)
		for (name, school) in schools.iteritems()
		if school.valid()
	)
	logging.debug('configured=%r', configured)

	for name in existing:
		if name not in configured:
			listener_remove(School(name))
			restart = True
	for school in configured.itervalues():
		if school.safe_name not in existing:
			listener_create(school)
			restart = True

	if restart:
		listener_restart()

		for name in existing:
			if name not in configured:
				listener_remove(School(name))


def listener_remove(school):
	"""
	Remove existing listener modules.
	"""
	logging.debug('Removing listener for "%s"...', school)
	_remove(school.listener_filename)
	_remove(school.status_filename)


def _remove(pathname):
	try:
		os.remove(pathname)
	except OSError as ex:
		if ex.errno != ENOENT:
			logging.warning('Failed to remove listener file "%s": %s', pathname, ex)


def listener_create(school):
	"""
	Clone new listener module.
	"""
	logging.info('Creating listener for "%s"...', school)
	tmp = school.listener_filename + '.tmp'
	with open(tmp, 'wb') as target:
		with open(TEMPLATE, 'rb') as template:
			for line in template:
				line = RE_SUBST.sub(school.safe_name, line)
				# lambda match: school.get(match.group(1), ''),
				target.write(line)
		os.fchmod(target.fileno(), 0755)
	os.rename(tmp, school.listener_filename)


def listener_restart():
	"""
	Restart Univention Directory Listener to pick up changes.
	"""
	cmd = ('/usr/sbin/invoke-rc.d', 'univention-directory-listener', 'restart')
	call(cmd)


def read_all(ucr):
	"""
	Read all configures schools from UCR.

	>>> from hashlib import md5
	>>> schools = read_all({
	...  'webweaver/UCS_WW3EE_Demo/ServerLogin': 'synchronisation@ucs.wwschool.de',
	...  'webweaver/UCS_WW3EE_Demo/ServerPasswd': md5('mj38zsASda'),
	...  'webweaver/UCS_WW3EE_Demo/error_email': 'true',
	... })
	>>> schools['UCS_WW3EE_Demo'].ServerLogin
	'synchronisation@ucs.wwschool.de'
	"""
	schools = {}
	for var, value in ucr.items():
		match = RE_UCR.match(var)
		if not match:
			continue
		school_name, key, subkey = match.groups()
		school = schools.setdefault(school_name, School(school_name))
		setattr(school, key, value)

	for school in schools.itervalues():
		school.fill_missing(ucr)

	return schools


if __name__ == '__main__':
	WebWeaver().main()
